import time
import math
import numpy as np
import functools

from kivy.lang import Builder
from plyer import gps, accelerometer
from kivy.app import App
from kivy.properties import StringProperty, NumericProperty, ObjectProperty
from kivy.clock import Clock, mainthread
from kivy.uix.boxlayout import BoxLayout
from kivy.graphics import Line

import projection as pr



accelerometer.enable()

kv = '''
<MainWindow>:
    orientation: 'vertical'
    drawing:drawing
    info:info
    
    Label:
        text:'drawing'
        id:drawing
        size_hint:1,1

    Label:
        text: root.gps_location
        size_hint:1,.3

    Label:
        id: info
        # text: root.info
        
    Label:
        id:accel_info
        text: root.accelerometer_data

    BoxLayout:
        size_hint_y: None
        height: '48dp'
        padding: '4dp'

        ToggleButton:
            text: 'Start' if self.state == 'normal' else 'Stop'
            on_state:
                root.start(root.min_time, root.min_distance) if self.state == 'down' else \
                root.stop()
                
        Button:
            text:'draw'
            on_release: root.on_draw()
            
        Button:
            text:'clear'
            on_release: root.canvas.clear()
'''

Builder.load_string(kv)

class MainWindow(BoxLayout):
    """"""
    min_distance=NumericProperty(0)
    min_time=NumericProperty(0)
    last_time=NumericProperty(0)
    diff_time=NumericProperty(0)
    ticker=ObjectProperty(None, allownone=True)

    gps_location = StringProperty()
    gps_status = StringProperty('Click Start to get GPS location updates')
    accelerometer_data = StringProperty("accal data coming here...")

    def __init__(self, *a, **kw):
        try:
            gps.configure(on_location=self.on_location,
                          on_status=self.on_status)
        except NotImplementedError:
            import traceback
            traceback.print_exc()
            self.gps_status = 'GPS is not implemented for your platform'
        print 'MainWindow'
        super(MainWindow, self).__init__(*a, **kw)

    def start(self, minTime, minDistance):
        try:
            gps.start(minTime, minDistance)
        except NotImplementedError:
            print 'GPS not implemented'
        self.ticker=Clock.schedule_interval(self.on_accel, .1)


    def stop(self):
        gps.stop()
        self.ticker.cancel()
        self.ticker=None

    @property
    def accel(self):
        try:
            accel=accelerometer.acceleration
        except Exception:
            accel=(0,1,0)

        return accel

    @mainthread
    def on_accel(self, *a, **kw):
        """"""

        normed=math.sqrt(sum(map(lambda x:x*x, self.accel)))
        self.accelerometer_data = "%1.4g\n%1.4g\n%1.4g\nabs:%2.5g" % (self.accel+(normed,))

        self.draw_cube()
        handy_mat=pr.make_cam_base(*self.accel)
        self.info.text=str(np.round(handy_mat,3))


    @mainthread
    def on_location(self, **kwargs):
        ts=time.time()
        self.diff_time=ts-self.last_time
        kwargs['diff_time']=self.diff_time
        print 'loc:', kwargs['lon'], kwargs['lat'], self.last_time, self.diff_time
        self.gps_location = '\n'.join([
            '{}={}'.format(k, v) for k, v in kwargs.items()])

        self.last_time=ts

    @mainthread
    def on_status(self, stype, status):
        """obsolete"""
        # self.gps_status = 'type={}\n{}'.format(stype, status)

    def on_pause(self):
        gps.stop()
        return True

    def on_resume(self):
        gps.start(self.min_time, self.min_distance)
        pass

    def line(self, x1,y1,x2,y2, **kw):
        with self.drawing.canvas:
            Line(points=[x1+self.drawing.x, y1+self.drawing.y,
                         x2+self.drawing.x,y2+self.drawing.y])

    def cube(self, sl):
        sl2 = sl / 2.
        corners = [pr.V(x, y, z) for x in [-sl2, sl2] for y in [-sl2, sl2] for z in [-sl2, sl2]]
        edges_cand = [(corner1, corner2)
                      for (i, corner1) in enumerate(corners)
                      for (j, corner2) in enumerate(corners[i:])]
        edges = filter(lambda (a, b): np.linalg.norm((a - b)) == sl, edges_cand)
        return corners, edges


    def draw_line(self, p1, p2, **kw):
        self.line(p1[0], p1[1], p2[0],  p2[1], **kw)


    def draw_edges(self, edges):

        for (p1, p2) in edges:
            self.draw_line(p1, p2)


    def make_proj_edges(self, PM):
        proj_point = lambda point: pr.unhom(PM * np.matrix(pr.hom(point)).T)[:2]
        proj_edge = lambda edge: map(proj_point, edge)
        proj_edges = lambda edges: map(proj_edge, edges)
        return proj_edges

    def make_plot_point_trafo(self, w,h):
        return lambda p: pr.unhom(pr.make_plot_matrix(w, h) * np.matrix(pr.hom(p)).T)

    def make_plot_trafo(self, w, h):
        point_trafo = self.make_plot_point_trafo(w,h)
        edge_trafo = lambda e: map(point_trafo, e)
        edges_trafo = lambda es: map(edge_trafo, es)
        #     return point_trafo
        return edges_trafo

    def on_draw(self):
        self.draw_cube()

    def draw_cube(self):
        points, edges = self.cube(4)
        ax,ay,az=self.accel
        Cam = pr.make_handy_cam(ax,ay,az, 7, 1)
        proj = self.make_proj_edges(Cam)
        ep=proj(edges)
        ep1=self.make_plot_trafo(self.drawing.width, self.drawing.height)(ep)

        self.drawing.canvas.clear()
        self.draw_edges(ep1)

class GpsTest(App):


    def build(self):

        return MainWindow()


if __name__ == '__main__':
    GpsTest().run()

