import numpy as np

def V(*d):
    return np.array(d)


def hom(v):
    """converts a vector to homogenous coords"""
    return np.concatenate([v,np.array([1])])


def unhom(v4):
    """converts a vector to unhomogenous coords"""
    if v4.ndim==2:
        v4=np.array(v4.T)[0]
    return (v4/v4[-1])[:-1]


def hom_mat(M3):
    """lifts a matrix into homogenous coordinates"""
    m1=np.insert(M3,M3.shape[1],0,axis=1)
    m2=np.insert(m1,m1.shape[0],0,axis=0)
    m2[-1,-1]=1
    return m2


def normalize(v):
    return v/np.linalg.norm(v)


def make_rot2(phi):
    """
    generates a rotation matrix
    :param phi: rotation angle in radians
    :return: numpy array of shape (2,2)
    """
    return np.matrix([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])


def lift_rot2(Rot, pos):
    liftRot2_ = lambda Rot, pos: np.insert(np.insert(Rot, pos, 0, axis=1), pos, 0, axis=0)
    R3 = liftRot2_(Rot, pos)
    R3[pos, pos] = 1
    return R3


def make_rot3(ax=0, ay=0, az=0):
    """
    creates a 3 x 3 rotation matrix given by rotation angles around the 3 axes
    :param ax: rotation angle around x-axis in radians
    :param ay: rotation angle around y-axis in radians
    :param az: rotation angle around z-axis in radians
    :return: numy array of shape (3,3)
    """
    rx = lift_rot2(make_rot2(ax), 0)
    ry = lift_rot2(make_rot2(ay), 1)
    rz = lift_rot2(make_rot2(az), 2)
    return rz * ry * rz

    return np.matmul(rz, np.matmul(ry, rx))


def make_rot3_deg(ax=0, ay=0, az=0):
    """
    creates a 3 x 3 rotation matrix given by rotation angles around the 3 axes
    :param ax: rotation angle around x-axis in radians
    :param ay: rotation angle around y-axis in radians
    :param az: rotation angle around z-axis in radians
    :return: numy array of shape (3,3)
    """
    return make_rot3(np.radians(ax), np.radians(ay), np.radians(az))


def make_proj(focal_length):
    return np.matrix(([1, 0, 0, 0],
                      [0, 1, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, -1. / focal_length, 0]), dtype='float64')


def make_trans(dx, dy=0, dz=0):
    return np.matrix([
        [1, 0, 0, dx],
        [0, 1, 0, dy],
        [0, 0, 1, dz],
        [0, 0, 0, 1]],
        dtype='float64')


def make_cam(R, dist):
    """
    :param R: a rotation matrix in R^3 that describes the CAMERA rotation
    :param dist: distance to origin
    :return: the P^4 matrix that has to be applied to the target object"""
    T = make_trans(0, 0, -dist)  # Move.subs(dx,0).subs(dy,0).subs(dz,-dist)
    # return np.matmul(T, hom_mat(R))
    return T * hom_mat(R)

# TODO: dynamisch kalibrieren

# correction matrix given by calibrating galaxy not 2014
MCorr = np.matrix([[0.998592366852551, 0.0530404078398720, 2.71432620675538e-20],
                   [-0.0530401588102975, 0.998587678369897, 0.00306433699154764],
                   [0.000162533683792965, -0.00306002352926993, 0.999995304908379]])


# Lenovo K6s
# MCorr=Matrix([[0.0755741681130987, 0.997140183281174, -2.98713477654379e-17],
#               [-1.60131332489823, 0.121365004085367, 2.32474000064970],
#               [2.31809167032889, -0.175690291628344, -1.60590592150144]])

# No Correction
MCorr=np.eye(3)


def make_cam_base(ax, ay, az):
    sqrt = np.sqrt
    a_x, a_y, a_z = normalize(V(ax, ay, az))

    D = np.matrix([[sqrt(-a_x**2 + 1), -a_x*a_y/sqrt(-a_x**2 + 1), -a_x*a_z/sqrt(-a_x**2 + 1)],
                 [a_x, a_y, a_z],
                 [0,
                  -a_x**2*a_z/sqrt(-a_x**2 + 1) - a_z*sqrt(-a_x**2 + 1),
                  a_x**2*a_y/sqrt(-a_x**2 + 1) + a_y*sqrt(-a_x**2 + 1)
                  ]
                 ])

    return D


def make_handy_cam(ax, ay, az, dist, f):
    D = make_cam_base(ax, ay, az)
    Rot = MCorr * D
    P = make_proj(f)
    C = make_cam(np.linalg.inv(Rot), dist)
    return P * C


def make_plot_matrix(w,h):
    """
    creates a matrix that converts NDC coords to screen coords
    :param w: the width of the output area
    :para h: the height
    :return: the 3 x 3 transformation matrix that takes a homogenized point in NDC"""
    T=np.matrix([[min(h, w)/2, 0, w/2], [0, min(h, w)/2, h/2], [0, 0, 1]])
    return T
